//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2019 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package auth

import (
	"context"
	"encoding/json"
	"net/http"

	oidc "github.com/coreos/go-oidc"
	"golang.org/x/oauth2"
)

type OIDCSession struct {
	State string
	Nonce string
}

func NewOIDCSession() (s *OIDCSession, err error) {
	s = &OIDCSession{}
	if s.State, err = generateRandomString(32); err != nil {
		return
	}
	if s.Nonce, err = generateRandomString(32); err != nil {
		return
	}
	return
}

// TODO: check concurrent usage of provider/verifier etc...

type OIDCBackend struct {
	provider     *oidc.Provider
	verifier     *oidc.IDTokenVerifier
	oauth2Config oauth2.Config
}

func NewOIDCBackend(cfg *OIDCConfig) (b *OIDCBackend, err error) {
	// TODO: make ctx a parameter?
	ctx := context.Background()
	b = &OIDCBackend{}
	if b.provider, err = oidc.NewProvider(ctx, cfg.IssuerURL); err != nil {
		return
	}

	oidcConfig := &oidc.Config{
		ClientID: cfg.ClientID,
	}
	b.verifier = b.provider.Verifier(oidcConfig)

	b.oauth2Config = oauth2.Config{
		ClientID:     cfg.ClientID,
		ClientSecret: cfg.ClientSecret,
		Endpoint:     b.provider.Endpoint(),
		RedirectURL:  cfg.CallbackURL,
		Scopes:       []string{oidc.ScopeOpenID, "username", "aura_shows"},
	}

	return
}

func (b *OIDCBackend) CallbackHandler() http.Handler {
	return &oidcCallbackHandler{backend: b}
}

func (b *OIDCBackend) HandleLogin(w http.ResponseWriter, r *http.Request) {
	var s *Session
	var sid string

	sc, err := r.Cookie(SessionCookieName)
	if sc != nil && err == nil {
		sid = sc.Value
		s = auth.sessions.get(sc.Value)
		if s != nil && s.Expired() {
			auth.sessions.remove(sid)
			s = nil
		}
	}

	if s != nil && s.oidc != nil {
		if s.Username != "" {
			http.Error(w, "you stilled logged in, please logout first!", http.StatusBadRequest)
			return
		}
		http.Error(w, "OIDC login already in progress, retry later", http.StatusConflict)
		return
	}

	if s == nil {
		s = &Session{}
		if s.oidc, err = NewOIDCSession(); err != nil {
			http.Error(w, err.Error(), http.StatusInternalServerError)
			return
		}
		if sid, err = auth.sessions.insert(s); err != nil {
			http.Error(w, err.Error(), http.StatusInternalServerError)
			return
		}

		// TODO: make cookie settings more secure!
		sc = &http.Cookie{Name: SessionCookieName, Value: sid, Expires: s.Expires}
		http.SetCookie(w, sc)
	}

	http.Redirect(w, r, b.oauth2Config.AuthCodeURL(s.oidc.State, oidc.Nonce(s.oidc.Nonce)), http.StatusFound)
}

type oidcCallbackHandler struct {
	backend *OIDCBackend
}

func invalidateSessionCookie(w http.ResponseWriter) {
	// TODO: this needs to improve
	http.SetCookie(w, &http.Cookie{Name: SessionCookieName, Value: "invalid", MaxAge: -1})
}

func (h *oidcCallbackHandler) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	sc, err := r.Cookie(SessionCookieName)
	if sc == nil || err != nil {
		http.Error(w, "no session cookie found, who are you?", http.StatusUnauthorized)
		return
	}
	s := auth.sessions.get(sc.Value)
	if s == nil {
		invalidateSessionCookie(w)
		http.Error(w, "invalid session cookie or session already expired", http.StatusUnauthorized)
		return
	}
	if s.oidc == nil {
		http.Error(w, "this not an OIDC session", http.StatusUnauthorized)
		return
	}
	if s.Username != "" {
		http.Error(w, "this session is already logged in", http.StatusConflict)
		return
	}

	if r.URL.Query().Get("state") != s.oidc.State {
		invalidateSessionCookie(w)
		auth.sessions.remove(sc.Value)
		http.Error(w, "OIDC: state did not match", http.StatusBadRequest)
		return
	}

	oauth2Token, err := h.backend.oauth2Config.Exchange(r.Context(), r.URL.Query().Get("code"))
	if err != nil {
		invalidateSessionCookie(w)
		auth.sessions.remove(sc.Value)
		http.Error(w, "OIDC: failed to exchange token: "+err.Error(), http.StatusBadRequest)
		return
	}

	rawIDToken, ok := oauth2Token.Extra("id_token").(string)
	if !ok {
		invalidateSessionCookie(w)
		auth.sessions.remove(sc.Value)
		http.Error(w, "OIDC: no id_token field in oauth2 token.", http.StatusInternalServerError)
		return
	}
	// Verify the ID Token signature and nonce.
	idToken, err := h.backend.verifier.Verify(r.Context(), rawIDToken)
	if err != nil {
		invalidateSessionCookie(w)
		auth.sessions.remove(sc.Value)
		http.Error(w, "OIDC: failed to verify ID Token: "+err.Error(), http.StatusInternalServerError)
		return
	}
	if idToken.Nonce != s.oidc.Nonce {
		invalidateSessionCookie(w)
		auth.sessions.remove(sc.Value)
		http.Error(w, "OIDC: invalid ID token nonce", http.StatusInternalServerError)
		return
	}

	userInfo, err := h.backend.provider.UserInfo(r.Context(), oauth2.StaticTokenSource(oauth2Token))
	if err != nil {
		invalidateSessionCookie(w)
		auth.sessions.remove(sc.Value)
		http.Error(w, "OIDC: failed to get userinfo: "+err.Error(), http.StatusInternalServerError)
		return
	}

	resp := struct {
		OAuth2Token *oauth2.Token
		UserInfo    *json.RawMessage
	}{oauth2Token, new(json.RawMessage)}

	if err := userInfo.Claims(&resp.UserInfo); err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	data, err := json.MarshalIndent(resp, "", "    ")
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
	w.Write(data)
}
